use std::{str::FromStr, path::Path, option::Option};

use itertools::Itertools;

use crate::{input::read_into_vec, error::AdventError};

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct Number {
    value: u64,
    position: u8,
    length: u8
}

#[derive(Debug, Clone)]
pub struct Symbol {
    position: u8,
    symbol: char,
}

impl From<u8> for Symbol {
    fn from(value: u8) -> Self {
        Symbol{position: value, symbol: ' '}
    }
}

impl PartialEq for Symbol {
    fn eq(&self, other: &Self) -> bool {
        self.position == other.position
    }
}

impl PartialOrd for Symbol {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        match self.position.partial_cmp(&other.position) {
            Some(core::cmp::Ordering::Equal) => {}
            ord => return ord,
        }
        self.symbol.partial_cmp(&other.symbol)
    }
}

impl Eq for Symbol {

}

impl Ord for Symbol {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.position.cmp(&other.position)
    }
}

impl FromStr for RowInfo {
    type Err = AdventError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut row_info = RowInfo::new();
        for (i, b) in s.as_bytes().into_iter().enumerate() {
            row_info.process(i, b)?;
        }
        Ok(row_info.finish(s.len())?.clone())
    }
}

#[derive(Debug, Clone)]
struct RowInfo {
    partial_start: Option<u8>,
    partial_number: Option<String>,
    pub symbols: Vec<Symbol>,
    pub numbers: Vec<Number>,
}

impl RowInfo {
    fn new() -> Self {
        Self {
            partial_start: None,
            partial_number: None,
            symbols: Vec::<Symbol>::new(),
            numbers: Vec::<Number>::new()
        }
    }

    fn add_symbol(&mut self, b: u8, i: usize) -> Result<(), AdventError> {
        self.symbols.push(Symbol {position: i as u8, symbol: b as char});
        if self.partial_start != None {
            self.finish_digit(i)?
        }
        Ok(())
    }

    fn finish_digit(&mut self, i: usize) -> Result<(), AdventError> {
        let number = Number {
            length: i as u8 - self.partial_start.ok_or(AdventError("Expected a partial to be started.".into()))?,
            position: self.partial_start.ok_or(AdventError("Expected a partial to be started.".into()))?,
            value: self.partial_number.clone().ok_or(AdventError("Expected partial to be started and have valid value.".into()))?.trim().parse()?
        };
        self.numbers.push(number);
        self.partial_start = None;
        self.partial_number = None;
        Ok(())
    }

    fn process(&mut self, i: usize, b: &u8) -> Result<(), AdventError> {
        if b != &b'.' && !b.is_ascii_digit() {
            self.add_symbol(*b, i)?
        }
        else if b.is_ascii_digit() {
            if self.partial_start.is_none() {
                self.partial_start = Some(i as u8);
                self.partial_number = Some((*b as char).to_string())
            } else {
                self.partial_number = Some(format!("{}{}", self.partial_number.as_ref().unwrap(), *b as char));
            }
        } else {
            if self.partial_start != None {
                self.finish_digit(i)?
            }
        }
        Ok(())
    }

    fn finish(&mut self, i: usize) -> Result<&Self, AdventError> {
        if self.partial_start != None {
            self.finish_digit(i)?;
        }
        return Ok(self);
    }
}

impl Default for RowInfo {
    fn default() -> Self {
        Self::new()
    }
}

pub fn gen_coords_number(y: usize, number: &Number) -> Vec<(u8, u8)> {
    let mut vec = Vec::<(u8, u8)>::new();
    let min_x = if number.position == 0 {
        0
    } else {
        number.position - 1
    };
    let max_x = number.position + number.length + 1;
    for x in min_x..max_x {
        if y != 0 {
            vec.push((x, y as u8 - 1));
        }
        if x < number.position || x >= number.position + number.length {
            vec.push((x, y as u8));
        }
        vec.push((x, y as u8 + 1));
    }
    vec
}

pub fn gen_coords_symbol(y: usize, symbol: &Symbol) -> Vec<(u8,u8)> {
    let mut vec = Vec::<(u8, u8)>::new();
    let min_x = if symbol.position == 0 {
        0
    } else {
        symbol.position - 1
    };
    let max_x = symbol.position + 1 + 1;
    for x in min_x..max_x {
        if y != 0 {
            vec.push((x, y as u8 - 1));
        }
        if x < symbol.position || x > symbol.position {
            vec.push((x, y as u8));
        }
        vec.push((x, y as u8 + 1));
    }
    vec
}

fn filter_gears(rows: &mut Vec<RowInfo>) -> &mut Vec<RowInfo> {
    for row in rows.iter_mut() {
        row.symbols = row.symbols.clone().into_iter().filter(|x| { x.symbol == '*' }).collect::<Vec<Symbol>>()
    }
    rows
}

pub fn one() -> Result<u64, AdventError> {
    let rows = read_into_vec::<RowInfo>(Path::new("resources/input3.txt"))?;
    let mut sum: u64 = 0;
    for (i, row) in rows.iter().enumerate() {
        for number in row.numbers.iter() {
            let coords = gen_coords_number(i, number);
            if coords.iter().any(|c| -> bool {
                if c.1 as usize >= rows.len() {
                    false
                } else {
                    rows[c.1 as usize].symbols.binary_search(&c.0.into()).is_ok()
                }
            }) {
                sum += number.value;
            }
        }
    }
    Ok(sum)
}

pub fn two() -> Result<u64, AdventError> {
    let mut rows = read_into_vec::<RowInfo>(Path::new("resources/input3.txt"))?;
    rows = filter_gears(&mut rows).clone();
    let mut sum: u64 = 0;
    for (i, row) in rows.iter().enumerate() {
        for symbol in row.symbols.iter() {
            let coords = gen_coords_symbol(i, symbol);
            let numbers = coords.iter().map(|(x, y)| {
                if *y as usize >= rows.len() {
                    None
                } else {
                    let relevant_numbers = rows[*y as usize].numbers.clone().into_iter().filter(|number| {*x >= number.position && *x < number.position + number.length}).collect::<Vec<Number>>();
                    if relevant_numbers.len() == 1 {
                        Some(relevant_numbers[0].clone())
                    } else {
                        None
                    }
                }
            })
            .filter(|x| {x.is_some()})
            .map(|x| {x.unwrap()})
            .unique()
            .collect::<Vec<Number>>();

            if numbers.len() == 2 {
                sum += numbers[0].value * numbers[1].value;
            }
        }
    }

    Ok(sum)
}