use std::{path::Path, str::FromStr, collections::HashMap};

use itertools::Itertools;
use num::integer::gcd;

use crate::{error::AdventError, input::read_into};

#[derive(Clone, Debug)]
enum Direction {
    Left,
    Right,
}

impl FromStr for Direction {
    type Err = AdventError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "L" => Ok(Direction::Left),
            "R" => Ok(Direction::Right),
            _ => Err(AdventError(format!("Unexpected input for Direction: {}", s))),
        }
    }
}

#[derive(Clone, Debug)]
struct Node {
    left: String,
    right: String
}

struct BaseMap {
    catalog: HashMap<String, Node>,
    directions: Vec<Direction>,
    start: Vec<String>,
}

trait Step {
    fn take_step(&self, current: &String, direction: &Direction) -> Result<String, AdventError>;
}

impl Step for BaseMap {
    fn take_step(&self, current: &String, direction: &Direction) -> Result<String, AdventError> {
        if let Some(node) = self.catalog.get(current) {
            Ok(match direction {
                Direction::Left => node.left.clone(),
                Direction::Right => node.right.clone()
            })
        }
        else {
            Err(AdventError(format!("Node missing in catalog for: {}", current)))
        }
    }
}

struct Map1(BaseMap);
struct Map2(BaseMap);

impl FromStr for BaseMap {
    type Err = AdventError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut lines = s.lines();
        let directions = &lines.next().unwrap().chars().map(|c| {
            Ok(Direction::from_str(format!("{}", c).as_str())?)
        }).collect::<Result<Vec<Direction>, AdventError>>()?;

        let mut nodes = HashMap::<String, Node>::new();
        for line in lines.skip(1) {
            let name = &line[0..3];
            let left = &line[7..10];
            let right = &line[12..15];

            nodes.insert(name.to_string(), Node{left: left.to_string(), right: right.to_string()});
        };

        Ok(BaseMap{
            catalog: nodes,
            directions: directions.to_vec(),
            start: vec![]
        })
    }
}

impl FromStr for Map1 {
    type Err = AdventError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut base_map = BaseMap::from_str(s)?;
        base_map.start.push("AAA".to_string());
        Ok(Map1(base_map))
    }
}

impl FromStr for Map2 {
    type Err = AdventError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut base_map = BaseMap::from_str(s)?;
        base_map.start = base_map.catalog.keys().filter(|x| x.ends_with('A')).map(|x| x.clone()).collect_vec();
        Ok(Map2(base_map))
    }
}

fn is_end1(node: &String) -> bool {
    node == &"ZZZ".to_string()
}

fn is_end2(node: &String) -> bool {
    node.ends_with('Z')
}

pub fn one() -> Result<u64, AdventError> {
    let map = read_into::<Map1>(Path::new("resources/input8.txt"))?.0;
    let mut node = map.start[0].clone();
    for (i, direction) in map.directions.iter().cycle().enumerate() {
        node = map.take_step(&node, direction)?;
        if is_end1(&node) {
            return Ok(i as u64 + 1);
        }
    }
    panic!("Cycle iterator should never terminate")
}

pub fn two() -> Result<u64, AdventError> {
    let map = read_into::<Map2>(Path::new("resources/input8.txt"))?.0;
    let mut nodes = map.start.clone();
    let iterations = nodes.iter_mut().map(|current| {
        for (i, direction) in map.directions.iter().cycle().enumerate() {
            current.clone_from(&map.take_step(current, direction)?);
            if is_end2(&current) {
                return Ok((i + 1) as u64);
            }  
        }
        panic!("Cycle iterator should never terminate")
    }).collect::<Result<Vec<u64>, AdventError>>()?;
    let mut iter = iterations.into_iter();
    let acc = iter.next().unwrap() as u64;
    Ok(iter.fold(Ok(acc), |acc, i| -> Result<u64, AdventError> {
        let accumulator = acc?;
        return Ok(i * accumulator / gcd(i, accumulator));
    })?)
}